import scala.concurrent.Future
import scala.util.{Random, Success}
import scala.concurrent.duration._

object Test {
  def main(args: Array[String]): Unit = args match {
    case Array("CvRDT",tpe) =>
      import CvRDT._
      tpe match {
        case "Counter" =>
          testCvRDT(Counter)(0, Counter.Inc(Random.nextInt(5)))
        case "PNCounter" =>
          testCvRDT(PNCounter)((0, 0),
            if (Random.nextBoolean()) PNCounter.Inc(Random.nextInt(5))
            else PNCounter.Dec(Random.nextInt(5))
          )
        case "GrowOnlySet" =>
          val GrowOnlyIntSet = new GrowOnlySet[Int]
          testCvRDT(GrowOnlyIntSet)(Set.empty, GrowOnlyIntSet.Add(Random.nextInt(1000)))
        case "TwoPhaseSet" =>
          val TwoPhaseIntSet = new TwoPhaseSet[Int]
          testCvRDT(TwoPhaseIntSet)((Set.empty, Set.empty), {
            if (Random.nextBoolean()) TwoPhaseIntSet.Add(Random.nextInt(1000))
            else TwoPhaseIntSet.Remove(Random.nextInt(1000))
          })
      }
    case Array("CmRDT",tpe) =>
      import CmRDT._
      tpe match {
        case "Counter" =>
          testCmRDT(Counter)(i => 0, (s: Int) => Counter.Inc(Random.nextInt(5)))
        case "LWWRegister" =>
          val LWWIntRegister = new LWWRegister[Int]
          testCmRDT(LWWIntRegister)(
            i => LWWIntRegister.RegisterState(i,Vector.fill(10)(0),0),
            s => LWWIntRegister.Write(s.time.updated(s.id,s.time(s.id) + 1),Random.nextInt(1000))
          )
      }
    case other =>
      println("usage: run (CvRDT|CmRDT) tpe")
  }

  def testCvRDT[T <: CvRDT](t: T)(initialState: t.State, randomOp: => t.Operation) = {
    import akka.actor._
    import akka.pattern.ask

    implicit val system = ActorSystem("test")
    implicit val timeout: akka.util.Timeout = akka.util.Timeout(1 second)
    implicit val executionContext: scala.concurrent.ExecutionContext = system.dispatcher

    case class Init(s: t.State, a: ActorRef)
    case object Freeze
    case object Move
    case object Talk
    case class Update(s: t.State)
    case object Query

    class Node extends Actor with ActorLogging {
      implicit def dispatcher = context.dispatcher

      def receive = {
        case Init(s, a) =>
          log.info(s"initialized with $s")
          context.become(initialized(s,a))
          context.system.scheduler.schedule(100 milliseconds, 100 milliseconds, self, Talk)
          context.system.scheduler.scheduleOnce(Random.nextInt(500).milliseconds, self, Move)
      }

      def initialized(s: t.State, other: ActorRef): Receive = {
        case Freeze =>
          log.info(s"freezing in state ${t.query(s)}")
          context.become(frozen(s,other))
        case Update(s2) =>
          if (s != s2) {
            val merged = t.merge(s, s2)
            log.info(s"merge: (${t.query(s)},${t.query(s2)}) -> ${t.query(merged)}")
            context.become(initialized(merged, other))
          }
        case Talk =>
          other ! Update(s)
        case Move =>
          val op = randomOp
          val newState = t.update(s,op)
          log.info(s"update: $op -> ${t.query(newState)}")
          context.become(initialized(newState,other))
          context.system.scheduler.scheduleOnce(Random.nextInt(500).milliseconds, self, Move)
        case Query => sender ! t.query(s)
      }

      def frozen(s: t.State, other: ActorRef): Receive = {
        case Update(s2) =>
          if (s != s2) {
            val merged = t.merge(s, s2)
            log.info(s"merge: ${t.query(s)} and ${t.query(s2)} to: ${t.query(merged)}")
            context.become(frozen(merged, other))
          }
        case Talk =>
          other ! Update(s)
        case Move =>
        case Query => sender ! t.query(s)
      }
    }

    val nodes = for (i <- 1 to 10) yield system.actorOf(Props(new Node),s"node$i")

    nodes.zip(nodes.drop(1) ++ nodes.take(1)).foreach {
      case (n,a) => n ! Init(initialState,a)
    }

    readLine()
    nodes.foreach(_ ! Freeze)

    readLine()
    Future.sequence(nodes.map(_ ? Query)).onComplete {
      case Success(s) =>
        s.zipWithIndex.foreach { case (v,i) => println(s"node ${i + 1}: $v") }
        system.terminate()
    }
  }

  def testCmRDT[T <: CmRDT](t: T)(initialState: Int => t.State, randomOp: t.State => t.Operation) = {
    import akka.actor._
    import akka.pattern.ask

    implicit val system = ActorSystem("test")
    implicit val timeout: akka.util.Timeout = akka.util.Timeout(1 second)
    implicit val executionContext: scala.concurrent.ExecutionContext = system.dispatcher

    case class Init(s: t.State, others: Seq[ActorRef])
    case object Freeze
    case object Move
    case class Update(owner: ActorRef, op: t.Operation)
    case object Query

    class Node extends Actor with ActorLogging {
      implicit def dispatcher = context.dispatcher

      def receive = {
        case Init(s, others) =>
          log.info(s"initialized with $s")
          context.become(initialized(s,others))
          context.system.scheduler.scheduleOnce(Random.nextInt(5000).milliseconds, self, Move)
      }

      def initialized(s: t.State, others: Seq[ActorRef]): Receive = {
        case Freeze =>
          log.info(s"freezing in state ${t.query(s)}")
          context.become(frozen(s, others))
        case Update(owner,op) =>
          if (owner != self) {
            val updated = t.updateExternal(s, op)
            log.info(s"updateExternal: (${t.query(s)},${op}) -> ${t.query(updated)}")
            context.become(initialized(updated, others))
            others.foreach(other =>
              context.system.scheduler.scheduleOnce(1000 milliseconds, other, Update(owner,op)))
          }
        case Move =>
          val op = randomOp(s)
          val newState = t.updateLocal(s,op)
          log.info(s"updateLocal: (${t.query(s)},$op) -> ${t.query(newState)}")
          context.become(initialized(newState,others))
          others.foreach(other => context.system.scheduler.scheduleOnce(1000 milliseconds, other, Update(self,op)))
          context.system.scheduler.scheduleOnce(Random.nextInt(5000).milliseconds, self, Move)
        case Query => sender ! t.query(s)
      }

      def frozen(s: t.State, others: Seq[ActorRef]): Receive = {
        case Update(owner,op) =>
          if (owner != self) {
            val updated = t.updateExternal(s, op)
            log.info(s"updateExternal: (${t.query(s)},${op}) -> ${t.query(updated)}")
            context.become(frozen(updated, others))
            others.foreach(other => context.system.scheduler.scheduleOnce(1000 milliseconds, other, Update(owner,op)))
          }
        case Move =>
        case Query => sender ! t.query(s)
      }
    }

    val nodes = for (i <- 1 to 10) yield system.actorOf(Props(new Node),s"node$i")

    nodes.zip(nodes.drop(1) ++ nodes.take(1)).zipWithIndex.foreach {
      case ((n,a),i) => n ! Init(initialState(i),Seq(a))
    }

    readLine()
    println("#################################################################################################")
    nodes.foreach(_ ! Freeze)

    readLine()
    println("#################################################################################################")
    Future.sequence(nodes.map(_ ? Query)).onComplete {
      case Success(s) =>
        s.zipWithIndex.foreach { case (v,i) => println(s"node ${i + 1}: $v") }
        system.terminate()
    }
  }
}